from __future__ import annotations

from typing import TYPE_CHECKING

from findpython.providers.base import BaseProvider

from pdm_conda.conda import conda_env_list

if TYPE_CHECKING:
    from pdm_conda.project import CondaProject
    from typing_extensions import Self
    from typing import Iterable
    from findpython.python import PythonVersion


class CondaProvider(BaseProvider):
    """A provider that finds python installed with Conda"""

    def __init__(self, project: CondaProject) -> None:
        super().__init__()
        self.project = project

    @classmethod
    def create(cls) -> Self | None:
        return None

    def find_pythons(self) -> Iterable[PythonVersion]:
        for env in conda_env_list(self.project):
            python_bin = env / "bin/python"
            if python_bin.exists():
                yield self.version_maker(python_bin, _interpreter=python_bin, keep_symlink=False)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.venv.create import CreateCommand as BaseCommand

from pdm_conda.models.config import CondaRunner

if TYPE_CHECKING:
    import argparse
    from pdm.project import Project


class CreateCommand(BaseCommand):
    description = BaseCommand.__doc__

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument(
            "-cn",
            "--conda-name",
            help="Specify the name of the Conda environment, overrides --name and appended hash",
        )

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        if conda_project := ((backend := (options.backend or project.config["venv.backend"])) in list(CondaRunner)):
            from pdm_conda.project import CondaProject

            project = cast(CondaProject, project)
            project.conda_config.runner = backend
            project.conda_config.is_initialized = True
            if options.conda_name:
                options.name = f"conda:{options.conda_name}"

        super().handle(project, options)

        # if inside a project ensure saving conda runner
        if conda_project:
            if project.pyproject.exists():
                project.pyproject.write(show_message=False)



from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from pdm.formats.base import array_of_inline_tables, make_array

from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement

if TYPE_CHECKING:
    from pdm.project import Project

    from pdm_conda.models.candidates import Candidate
    from pdm_conda.models.requirements import Requirement

_patched = False


def remove_quotes(req: str) -> str:
    for quote in ("'", '"'):
        if req.startswith(quote) and req.endswith(quote):
            req = req[1:-1]
    return req


def wrap_save_version_specifiers(func):
    @functools.wraps(func)
    def wrapper(
        requirements: dict[str, dict[str, Requirement]],
        resolved: dict[str, Candidate],
        save_strategy: str,
    ) -> None:
        func(requirements, resolved, save_strategy)
        for reqs in requirements.values():
            for name in reqs:
                if isinstance(can := resolved[name], CondaCandidate):
                    req = as_conda_requirement(reqs[name])
                    req.version_mapping.update(can.req.version_mapping)
                    req.is_python_package = can.req.is_python_package
                    reqs[name] = req

    return wrapper


def wrap_format_lockfile(func):
    @functools.wraps(func)
    def wrapper(
        project: Project,
        mapping: dict[str, Candidate],
        fetched_dependencies: dict[tuple[str, str | None], list[Requirement]],
        *args,
        **kwargs,
    ) -> dict:
        res = func(project, mapping, fetched_dependencies, *args, **kwargs)
        # ensure no duplicated groups in metadata
        if groups := res.get("metadata", dict()).get("groups"):
            res["metadata"]["groups"] = list({group: None for group in groups}.keys())

        assert len(res["package"]) == len(mapping)
        # fix conda packages
        for package, (_, can) in zip(res["package"], sorted(mapping.items())):
            # only static-url allowed for conda packages
            if isinstance(can, CondaCandidate):
                package["files"] = array_of_inline_tables(
                    [{"url": item["url"], "hash": item["hash"]} for item in can.hashes],
                    multiline=True,
                )

            # fix conda dependencies to include build string
            dependencies = []
            include_dependencies = False
            for dep in fetched_dependencies.get(can.dep_key, []):
                kwargs = {}
                if isinstance(dep, CondaRequirement):
                    kwargs["with_build_string"] = True
                    include_dependencies = True
                dependencies.append(dep.as_line(**kwargs))
            if include_dependencies:
                package["dependencies"] = make_array(sorted(dependencies), True)

        res["package"] = sorted(res["package"], key=lambda x: x["name"])
        return res

    return wrapper


if not _patched:
    from pdm.cli import actions, utils

    save_version_specifiers = wrap_save_version_specifiers(utils.save_version_specifiers)
    format_lockfile = wrap_format_lockfile(utils.format_lockfile)
    for m in [utils, actions]:
        setattr(m, "save_version_specifiers", save_version_specifiers)
        setattr(m, "format_lockfile", format_lockfile)
    _patched = True



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, cast

from pdm.installers import Synchronizer

from pdm_conda.environments import CondaEnvironment
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import strip_extras
from pdm_conda.models.setup import CondaSetupDistribution

if TYPE_CHECKING:
    from typing import Collection

    from pdm_conda.environments import BaseEnvironment
    from pdm_conda.models.candidates import Candidate


class CondaSynchronizer(Synchronizer):
    def __init__(
        self,
        candidates: dict[str, Candidate],
        environment: BaseEnvironment,
        clean: bool = False,
        dry_run: bool = False,
        retry_times: int = 1,
        install_self: bool = False,
        no_editable: bool | Collection[str] = False,
        reinstall: bool = False,
        only_keep: bool = False,
        fail_fast: bool = False,
    ) -> None:
        super().__init__(
            candidates,
            environment,
            clean,
            dry_run,
            retry_times,
            install_self,
            no_editable,
            reinstall,
            only_keep,
            fail_fast,
        )
        self.environment = cast(CondaEnvironment, environment)
        self.parallel = bool(self.parallel)  # type: ignore

    @cached_property
    def candidates(self) -> dict[str, Candidate]:
        candidates = super().candidates
        # if key is requirement with extras, add candidate without extras if it doesn't exist
        for key in list(candidates):
            if (
                isinstance(can := candidates[key], CondaCandidate)
                and candidates.get(name := strip_extras(key)[0], None) is None
            ):
                candidates[name] = can
        return candidates

    def compare_with_working_set(self) -> tuple[list[str], list[str], list[str]]:
        to_add, to_update, to_remove = super().compare_with_working_set()
        if not isinstance(self.environment, CondaEnvironment):
            return to_add, to_update, to_remove

        # deactivate parallel execution if uninstall
        self.parallel = self.environment.project.config["install.parallel"]
        if to_remove:
            to_remove = [p for p in to_remove if p not in self.environment.env_dependencies]

        num_install = 0
        for pkgs in (to_add, to_update):
            num_install += len([p for p in pkgs if isinstance(self.candidates[p], CondaCandidate)])

        num_remove = 0
        for pkgs in (to_remove, to_update):
            num_remove += len([p for p in pkgs if isinstance(self.working_set[p], CondaSetupDistribution)])

        if self.parallel and num_remove > 0:
            if not self.dry_run:
                self.environment.project.core.ui.echo("Deactivating parallel uninstall.")
            self.parallel = False

        if self.environment.project.conda_config.batched_commands:
            self.manager.prepare_batch_operations(
                num_install=num_install,
                num_remove=num_remove,
            )

        return to_add, to_update, to_remove



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, cast

from pdm import termui
from pdm.exceptions import CandidateNotFound
from pdm.models.repositories import BaseRepository, LockedRepository, PyPIRepository
from pdm.models.specifiers import PySpecSet

from pdm_conda import logger
from pdm_conda.conda import (
    CondaResolutionError,
    CondaSearchError,
    conda_create,
    conda_search,
    sort_candidates,
)
from pdm_conda.environments import CondaEnvironment
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement

if TYPE_CHECKING:
    from typing import Any, Iterable, Mapping

    from pdm.models.repositories import CandidateKey, RepositoryConfig

    from pdm_conda.environments import BaseEnvironment
    from pdm_conda.models.candidates import Candidate, FileHash
    from pdm_conda.models.requirements import Requirement


class CondaRepository(BaseRepository):
    def __init__(
        self,
        sources: list[RepositoryConfig],
        environment: BaseEnvironment,
        ignore_compatibility: bool = True,
    ) -> None:
        super().__init__(sources, environment, ignore_compatibility)
        self.environment = cast(CondaEnvironment, environment)
        self._conda_resolution: dict[str, list[CondaCandidate]] = dict()

    def is_conda_managed(self, requirement: Requirement) -> bool:
        """
        True if requirement is conda requirement or (not excluded and named requirement
        and conda as default manager or used by another conda requirement)
        :param requirement: requirement to evaluate
        """
        if not isinstance(self.environment, CondaEnvironment):
            return False
        from pdm_conda.models.requirements import is_conda_managed as _is_conda_managed

        return _is_conda_managed(requirement, self.environment.project.conda_config)

    def update_conda_resolution(
        self,
        requirements: list[Requirement] | None = None,
        resolution: dict | None = None,
    ):
        """
        Updates the existing conda resolution if new requirements.
        :param requirements: list of requirements
        :param resolution: resolution to override existing
        :return: list of changed requirements
        """
        if resolution is not None:
            self._conda_resolution = resolution

    def get_dependencies(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
        if isinstance(candidate, CondaCandidate):
            dependencies = list(candidate.dependencies)
            # if dep in constrains use constrain
            if candidate.constrains:
                for i, dep in enumerate(dependencies):
                    if (constrain := candidate.constrains.get(dep.identify(), None)) is not None:
                        dependencies[i] = constrain
            requires_python = PySpecSet(candidate.requires_python)
            summary = candidate.summary
        else:
            dependencies, requires_python, summary = super().get_dependencies(candidate)
        return dependencies, requires_python, summary

    def get_hashes(self, candidate: Candidate) -> list[FileHash]:
        if isinstance(candidate, CondaCandidate):
            if not candidate.hashes:
                termui.logger.info("Fetching hashes for %s", candidate)
                _candidates = conda_search(self.environment.project, candidate.req)
                if not _candidates:
                    raise CondaSearchError(f"Cannot find hashes for {candidate}")

                candidate.hashes = _candidates[0].hashes
        return super().get_hashes(candidate)


class PyPICondaRepository(PyPIRepository, CondaRepository):
    def update_conda_resolution(
        self,
        requirements: list[Requirement] | None = None,
        resolution: dict | None = None,
    ):
        super().update_conda_resolution(requirements, resolution)
        requirements = requirements or []
        requirements = [as_conda_requirement(req) for req in requirements if self.is_conda_managed(req)]
        update = False
        # if any requirement not in saved candidates or incompatible candidates
        for req in requirements:
            if update:
                break
            key = req.conda_name
            if key not in self._conda_resolution:
                update = True
                break
            for can in self._conda_resolution[key]:
                if not req.is_compatible(can):
                    update = True
                    break

        if update:
            try:
                resolution = conda_create(
                    self.environment.project,
                    requirements,
                    prefix=f"/tmp/{uuid.uuid4()}",
                    dry_run=True,
                )
                _requirements = {r.conda_name: r for r in requirements}
                for name, candidates in resolution.items():
                    req = _requirements.get(name, candidates[0].req)
                    key = req.conda_name
                    self._conda_resolution[key] = candidates
            except CondaResolutionError as err:
                logger.info(err)
                if err.packages:
                    msg = "Unable to find candidates for "
                    for i, package in enumerate(err.packages):
                        msg += f"[success]{package}[/success]"
                        if i < len(err.packages) - 1:
                            if i == len(err.packages) - 2:
                                msg += " and "
                            else:
                                msg += ", "
                    msg += " with Conda.\nYou should add more channels or add the packages to the excludes list."
                    raise CandidateNotFound(msg)
                raise

    def _find_candidates(self, requirement: Requirement, minimal_version: bool) -> Iterable[Candidate]:
        if self.is_conda_managed(requirement):
            requirement = as_conda_requirement(requirement)
            candidates = self._conda_resolution.get(requirement.conda_name, [])
            candidates = [
                c.copy_with(requirement, merge_requirements=True) for c in candidates if requirement.is_compatible(c)
            ]
            candidates = list(sort_candidates(self.environment.project, candidates, minimal_version))
        else:
            if isinstance(requirement, CondaRequirement):
                requirement = requirement.as_named_requirement()
            candidates = super()._find_candidates(requirement, minimal_version)
        return candidates


class LockedCondaRepository(LockedRepository, CondaRepository):
    def _matching_keys(self, requirement: Requirement) -> Iterable[CandidateKey]:
        yield from super()._matching_keys(requirement)
        if self.is_conda_managed(requirement):
            req_id = as_conda_requirement(requirement).identify()

            for key, can in self.packages.items():
                if isinstance(can, CondaCandidate) and req_id == key[0]:
                    yield key

    def _read_lockfile(self, lockfile: Mapping[str, Any]) -> None:
        packages = lockfile.get("package", [])
        conda_packages = []
        pypi_packages = []
        for package in packages:
            if package.get("conda_managed", False):
                conda_packages.append(package)
            else:
                pypi_packages.append(package)
        super()._read_lockfile({"package": pypi_packages, **{k: v for k, v in lockfile.items() if k != "package"}})

        for package in conda_packages:
            can = CondaCandidate.from_lock_package(package)
            can_id = self._identify_candidate(can)
            self.packages[can_id] = can
            self.candidate_info[can_id] = (
                can.dependencies_lines,
                package.get("requires_python", ""),
                package.get("summary", ""),
            )

    def _identify_candidate(self, candidate: Candidate) -> tuple:
        if isinstance(candidate, CondaCandidate):
            return candidate.identify(), candidate.version, None, False
        return super()._identify_candidate(candidate)



import hashlib
import json

from pdm.project.project_file import PyProject as PyProjectBase


class PyProject(PyProjectBase):
    def content_hash(self, algo: str = "sha256") -> str:
        """
        Generate a hash of the sensible content of the pyproject.toml file.
        When the hash changes, it means the project needs to be relocked.
        :param algo: hash algorithm name
        :return: pyproject.toml hash
        """
        pdm_conda_data = self.settings.get("conda", dict())
        pdm_conda_dump_data = dict()
        for hash_config in (
            "channels",
            "as-default-manager",
            "excludes",
            "dependencies",
            "dev-dependencies",
            "optional-dependencies",
        ):
            if hash_config in pdm_conda_data:
                pdm_conda_dump_data[hash_config] = pdm_conda_data[hash_config]
        if not pdm_conda_dump_data:
            return super().content_hash(algo)

        dump_data = {
            "sources": self.settings.get("source", []),
            "dependencies": self.metadata.get("dependencies", []),
            "dev-dependencies": self.settings.get("dev-dependencies", {}),
            "optional-dependencies": self.metadata.get("optional-dependencies", {}),
            "requires-python": self.metadata.get("requires-python", ""),
            "pdm-conda": pdm_conda_dump_data,
            "overrides": self.resolution.get("overrides", {}),
        }
        pyproject_content = json.dumps(dump_data, sort_keys=True)
        hasher = hashlib.new(algo)
        hasher.update(pyproject_content.encode("utf-8"))
        return hasher.hexdigest()



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.installers import InstallManager

from pdm_conda.conda import conda_install, conda_uninstall
from pdm_conda.environments import CondaEnvironment
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.setup import CondaSetupDistribution

if TYPE_CHECKING:
    from importlib.metadata import Distribution

    from pdm_conda.environments import BaseEnvironment
    from pdm_conda.models.candidates import Candidate


class CondaInstallManager(InstallManager):
    def __init__(self, environment: BaseEnvironment, *, use_install_cache: bool = False) -> None:
        super().__init__(environment, use_install_cache=use_install_cache)
        self.environment = cast(CondaEnvironment, environment)
        self._num_install = 0
        self._num_remove = 0
        self._batch_install: list[str] = []
        self._batch_remove: list[str] = []

    def prepare_batch_operations(self, num_install: int, num_remove: int):
        self._num_install = num_install
        self._num_remove = num_remove

    def _run_with_conda(self, conda_func, new_requirement: str, requirements: list[str], min_requirements: int):
        requirements.append(new_requirement)
        if len(requirements) >= min_requirements:
            try:
                conda_func(
                    self.environment.project,
                    list(requirements),
                    no_deps=True,
                )
                requirements.clear()
            except:
                if min_requirements == 0:
                    requirements.clear()
                raise

    def install(self, candidate: Candidate) -> Distribution:
        """
        Install candidate, use conda if conda package else default installer
        :param candidate: candidate to install
        """
        if isinstance(candidate, CondaCandidate):
            self._run_with_conda(
                conda_install,
                f"{candidate.link.url_without_fragment}#{candidate.link.hash}",
                self._batch_install,
                self._num_install,
            )
            return candidate.distribution
        else:
            return super().install(candidate)

    def uninstall(self, dist: Distribution) -> None:
        """
        Uninstall distribution, use conda if conda package else default uninstaller
        :param dist: distribution to uninstall
        """
        if isinstance(dist, CondaSetupDistribution):
            self._run_with_conda(conda_uninstall, dist.name, self._batch_remove, self._num_remove)
        else:
            super().uninstall(dist)

    def overwrite(self, dist: Distribution, candidate: Candidate) -> None:
        """
        Overwrite distribution with candidate, uninstall and install with conda if conda package else default overwrite
        :param dist: distribution to uninstall
        :param candidate: candidate to install
        """
        if isinstance(candidate, CondaCandidate) or isinstance(dist, CondaSetupDistribution):
            self.uninstall(dist)
            self.install(candidate)
        else:
            super().overwrite(dist, candidate)



from __future__ import annotations

import re
from typing import Iterable


class ChannelSorter:
    def __init__(self, platform: str, channels: Iterable[str] | None = None) -> None:
        self._priority: dict[str, int] = dict()
        self._tree: dict[str, list[str]] = dict()
        self.platform = platform
        if channels:
            for channel in channels:
                self.add_channel(channel)
            for root in self._tree:
                self.add_defaults(root)

    def get_root(self, channel: str) -> str:
        return channel.split("/")[0]

    def add_defaults(self, root: str):
        for channel in [f"{root}/{self.platform}", rf"{root}/.*", f"{root}/noarch"]:
            self.add_channel(channel, allow_fuzzy=not channel.endswith("noarch"))

    def get_variants(self, root: str):
        # add parent channel priority
        if root not in self._tree:
            self._priority[root] = 1000 * len(self._tree)
            self._tree[root] = []

        return self._tree[root]

    def add_channel(self, channel: str, allow_fuzzy=True):
        root = self.get_root(channel)
        if channel not in self._priority:
            for c in (variants := self.get_variants(root)):
                if c == channel or (allow_fuzzy and re.match(c, channel)):
                    self._priority[channel] = self._priority[c]
                    # then fuzzy match
                    if c != channel:
                        self._priority[c] += 1
                    break
            # couldn't find priority in saved variant
            if channel not in self._priority:
                self._priority[channel] = self._priority[root] + len(variants) * 10
                variants.append(channel)

    def get_priority(self, channel: str) -> int:
        self.add_channel(channel)
        return self._priority[channel]



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, cast

from pdm.exceptions import ProjectError
from pdm.project import Project
from pdm.project.lockfile import Lockfile
from pdm.utils import get_venv_like_prefix
from tomlkit.items import Array

from pdm_conda.models.config import PluginConfig
from pdm_conda.models.requirements import (
    CondaRequirement,
    as_conda_requirement,
    is_conda_managed,
    parse_requirement,
)
from pdm_conda.project.project_file import PyProject

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Iterable

    from findpython import Finder
    from pdm.core import Core
    from pdm.environments import BaseEnvironment
    from pdm.models.repositories import LockedRepository
    from pdm.models.requirements import Requirement
    from pdm.resolver.providers import BaseProvider


class CondaProject(Project):
    def __init__(
        self,
        core: Core,
        root_path: str | Path | None,
        is_global: bool = False,
        global_config: str | Path | None = None,
    ) -> None:
        from pdm_conda.environments import CondaEnvironment
        from pdm_conda.installers.manager import CondaInstallManager
        from pdm_conda.installers.synchronizers import CondaSynchronizer
        from pdm_conda.models.repositories import (
            LockedCondaRepository,
            PyPICondaRepository,
        )
        from pdm_conda.resolvers import CondaResolver

        super().__init__(core, root_path, is_global, global_config)
        self.core.repository_class = PyPICondaRepository
        self.core.install_manager_class = CondaInstallManager
        self.core.synchronizer_class = CondaSynchronizer
        self.core.resolver_class = CondaResolver
        self.locked_repository_class = LockedCondaRepository
        self.environment_class = CondaEnvironment
        self._virtual_packages: set[CondaRequirement] | None = None
        self._platform: str | None = None
        self._default_channels: list[str] | None = None
        self._conda_mapping: dict[str, str] = dict()
        self._pypi_mapping: dict[str, str] = dict()
        self.conda_config = PluginConfig.load_config(self)
        self._is_distribution: bool | None = None

    def _check_update_info(self, prop):
        if prop is None:
            self._get_conda_info()

    @cached_property
    def virtual_packages(self) -> set[CondaRequirement]:
        self._check_update_info(self._virtual_packages)
        return self._virtual_packages  # type: ignore

    @property
    def platform(self) -> str:
        self._check_update_info(self._platform)
        return self._platform  # type: ignore

    @property
    def default_channels(self) -> list[str]:
        self._check_update_info(self._default_channels)
        return self._default_channels  # type: ignore

    @property
    def locked_repository(self) -> LockedRepository:
        try:
            lockfile = self.lockfile._data.unwrap()
        except ProjectError:
            lockfile = {}

        return self.locked_repository_class(lockfile, self.sources, self.environment)

    @cached_property
    def pyproject(self) -> PyProject:
        return PyProject(self.root / self.PYPROJECT_FILENAME, ui=self.core.ui)

    @property
    def lockfile(self) -> Lockfile:
        if self._lockfile is None:
            self.set_lockfile(self.root / self.LOCKFILE_FILENAME)
        return self._lockfile

    def set_lockfile(self, path: str | Path) -> None:
        self._lockfile = Lockfile(path, ui=self.core.ui)
        # conda don't produce cross-platform locks
        if self.conda_config.is_initialized and not self._lockfile.empty():
            self._lockfile._data.setdefault("metadata", {})["cross_platform"] = False

    def _get_conda_info(self):
        from pdm_conda.conda import conda_info

        info = conda_info(self)
        self._virtual_packages = info["virtual_packages"]
        self._platform = info["platform"]
        self._default_channels = info["channels"]

    def get_conda_pyproject_dependencies(self, group: str, dev: bool = False, set_defaults=False) -> list[str]:
        """
        Get the conda dependencies array in the pyproject.toml
        """

        def _getter(conf, name, default, set_defaults=False):
            return (conf.setdefault if set_defaults else conf.get)(name, default)

        settings = _getter(self.pyproject.settings, "conda", dict(), set_defaults)
        if group == "default":
            group = "dependencies"
        else:
            name = "optional" if not dev else "dev"
            settings = _getter(settings, f"{name}-dependencies", dict(), set_defaults)
        return _getter(settings, group, [], set_defaults)

    def iter_groups(self, dev: bool = True) -> Iterable[str]:
        groups = set(super().iter_groups())
        config = self.conda_config
        for is_dev, deps in ((False, config.optional_dependencies), (True, config.dev_dependencies)):
            if deps and (dev or not is_dev):
                groups.update(deps.keys())
        if not dev:
            for group in self.pyproject.settings.get("dev-dependencies", {}):
                groups.remove(group)
        return groups

    def get_dependencies(self, group: str | None = None) -> dict[str, Requirement]:
        if group in super().iter_groups():
            result = super().get_dependencies(group)
        else:
            result = dict()

        config = self.conda_config
        group = group or "default"

        dev = group not in config.optional_dependencies
        if group in config.optional_dependencies and group in config.dev_dependencies:
            self.core.ui.echo(
                f"The {group} group exists in both [optional-dependencies] "
                "and [dev-dependencies], the former is taken.",
                err=True,
                style="warning",
            )
        deps = self.get_conda_pyproject_dependencies(group, dev)

        for line in deps:
            req = parse_requirement(f"conda:{line}")
            req.groups = [group]
            # search for package with extras to remove it
            pypi_req = next((v for v in result.values() if v.conda_name == req.conda_name), None)
            if pypi_req is not None:
                result.pop(pypi_req.identify())
                if not req.specifier:
                    req.specifier = pypi_req.specifier
                if pypi_req.marker:
                    req.marker = pypi_req.marker
                if pypi_req.extras:
                    req.extras = pypi_req.extras
                req.groups = pypi_req.groups
            result[req.identify()] = req

        if self.conda_config.as_default_manager:
            for k in list(result):
                if is_conda_managed(req := result[k], config):
                    result[k] = as_conda_requirement(req)

        return result

    def add_dependencies(
        self,
        requirements: dict[str, Requirement],
        to_group: str = "default",
        dev: bool = False,
        show_message: bool = True,
    ) -> None:
        conda_requirements = {n: r for n, r in requirements.items() if isinstance(r, CondaRequirement)}
        requirements = {n: r for n, r in requirements.items() if n not in conda_requirements} | {
            n: r.as_named_requirement() for n, r in conda_requirements.items() if r.is_python_package
        }
        if self.conda_config.as_default_manager:
            conda_requirements = {
                n: r for n, r in conda_requirements.items() if not r.is_python_package or r.channel or r.build_string
            }
        if conda_requirements:
            deps = self.get_conda_pyproject_dependencies(to_group, dev, set_defaults=True)
            cast(Array, deps).multiline(True)
            for _, dep in conda_requirements.items():
                matched_index = next((i for i, r in enumerate(deps) if dep.matches(f"conda:{r}")), None)
                req = dep.as_line(with_channel=True)
                if matched_index is None:
                    deps.append(req)
                else:
                    deps[matched_index] = req

        super().add_dependencies(requirements, to_group, dev, show_message)

    def get_environment(self) -> BaseEnvironment:
        if not self.conda_config.is_initialized:
            return super().get_environment()

        if not self.config["python.use_venv"]:
            raise ProjectError("python.use_venv is required to use Conda.")
        if get_venv_like_prefix(self.python.executable) is None:
            raise ProjectError("Conda environment not detected.")

        return self.environment_class(self)

    def get_provider(
        self,
        strategy: str = "all",
        tracked_names: Iterable[str] | None = None,
        for_install: bool = False,
        ignore_compatibility: bool = True,
        direct_minimal_versions: bool = False,
    ) -> BaseProvider:
        from pdm_conda.resolver.providers import BaseProvider, CondaBaseProvider

        kwargs = dict(direct_minimal_versions=direct_minimal_versions)
        provider = super().get_provider(strategy, tracked_names, for_install, ignore_compatibility, **kwargs)
        if isinstance(provider, BaseProvider) and not isinstance(provider, CondaBaseProvider):
            kwargs["locked_candidates"] = provider.locked_candidates
            return CondaBaseProvider(provider.repository, **kwargs)  # type: ignore[arg-type]
        return provider

    def _get_python_finder(self, search_venv: bool = True) -> Finder:
        finder = super()._get_python_finder(search_venv)
        from pdm_conda.cli.commands.venv.utils import CondaProvider

        if self.conda_config.is_initialized:
            finder.add_provider(CondaProvider(self), 0)
        return finder

    @property
    def is_distribution(self) -> bool:
        if self._is_distribution is not None:
            return self._is_distribution
        return super().is_distribution

    @is_distribution.setter
    def is_distribution(self, value: bool | None):
        self._is_distribution = value



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.add import Command as BaseCommand
from pdm.cli.options import ArgumentGroup
from pdm.exceptions import RequirementError

from pdm_conda.cli.utils import remove_quotes
from pdm_conda.models.requirements import (
    CondaRequirement,
    is_conda_managed,
    parse_requirement,
)
from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project

conda_group = ArgumentGroup("Conda Arguments")
conda_group.add_argument(
    "--conda",
    dest="conda_packages",
    action="append",
    help="Specify Conda packages",
    default=[],
)

conda_group.add_argument(
    "-c",
    "--channel",
    dest="conda_channel",
    type=str,
    help="Specify Conda channel",
    default="",
)

conda_group.add_argument(
    "-r",
    "--runner",
    dest="conda_runner",
    type=str,
    help="Specify Conda runner executable",
    default="",
)
conda_group.add_argument(
    "--conda-as-default-manager",
    dest="conda_as_default_manager",
    default=False,
    action="store_true",
    help="Specify Conda as default manager",
)
conda_group.add_argument(
    "-ce",
    "--conda-excludes",
    dest="conda_excludes",
    type=str,
    help="Specify Conda excluded dependencies separated by comma",
    default="",
)


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "add"
    arguments = (*BaseCommand.arguments, conda_group)

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        project = cast(CondaProject, project)
        config = project.conda_config
        if options.conda_runner:
            config.runner = options.conda_runner
        existing_channels = config.channels
        if (channel := options.conda_channel) and channel not in existing_channels:
            existing_channels.append(channel)
            config.channels = existing_channels
        if conda_excludes := options.conda_excludes:
            conda_excludes = set(conda_excludes.split(","))
            config.excludes = set(conda_excludes).union(config.excludes)
        if options.conda_as_default_manager:
            config.as_default_manager = True

        conda_packages = options.conda_packages
        for i, p in enumerate(conda_packages):
            if not p.startswith("conda:"):
                conda_packages[i] = f"conda:{remove_quotes(p)}"
        if config.as_default_manager:
            conda_packages += options.packages
            options.packages = []
        for package in conda_packages:
            package_channel = None
            package = remove_quotes(package)
            conda_package = package.startswith("conda:")

            if "::" in package:
                package_channel, package = package.split("conda:", maxsplit=1)[-1].split("::", maxsplit=1)
            _p = None

            try:
                _p = parse_requirement(package)
                if isinstance(_p, CondaRequirement):
                    _p = _p.as_named_requirement()
            except RequirementError:
                # if requirement error it can have an unparsable version
                pass

            # if not named we can't use Conda
            if conda_package or _p is None or is_conda_managed(_p, config):
                if package.startswith("conda:"):
                    package = package[len("conda:") :]
                if not package_channel and channel:
                    package_channel = channel

                if package_channel:
                    package = f"{package_channel}::{package}"
                    if package_channel not in existing_channels:
                        project.core.ui.echo(f"Detected Conda channel {package_channel}, adding it to pyproject")
                        existing_channels.append(package_channel)
                        config.channels = existing_channels

                package = f"conda:{package}"

            options.packages.append(package)

        super().handle(project, options)



from __future__ import annotations

from typing import TYPE_CHECKING

from pdm import termui

if TYPE_CHECKING:
    from pdm.core import Core

logger = termui.logger
__version__ = "0.16.5"


def main(core: Core):
    from pdm_conda import utils
    from pdm_conda.cli import utils as cli_utils
    from pdm_conda.cli.commands.add import Command as AddCommand
    from pdm_conda.cli.commands.init import Command as InitCommand
    from pdm_conda.cli.commands.install import Command as InstallCommand
    from pdm_conda.cli.commands.lock import Command as LockCommand
    from pdm_conda.cli.commands.remove import Command as RemoveCommand
    from pdm_conda.cli.commands.update import Command as UpdateCommand
    from pdm_conda.cli.commands.venv import Command as VenvCommand
    from pdm_conda.cli.commands.venv import backends
    from pdm_conda.environments import python
    from pdm_conda.models.config import CONFIGS
    from pdm_conda.project import CondaProject

    core.project_class = CondaProject

    for cmd in [InitCommand, AddCommand, RemoveCommand, UpdateCommand, LockCommand, InstallCommand, VenvCommand]:
        core.register_command(cmd)

    for name, config in CONFIGS:
        core.add_config(name, config)
